#include "cuik-kdtree.h"

#include "definitions.h"
#include "vector.h"
#include "random.h"

#include <assert.h>
#include <stdio.h>
#include <string.h>

/** 
   \file cuik-kdtree.c

   \brief Implementation of a kd-tree.

   Implementation of a kd-tree. This is used to speed up the search
   of nearest neighbours (or neighbours in a given ball) but also for
   sampling.

   This is an alternative to the DNN library, but here we use the
   kd-tree not only to search but also to sample. The DNN kd-tree
   is actually a forest of trees. To samle a single tree is
   necessary. So,instead of modifying the DNN implementation, 
   we implemented a new kd-tree, even at the risk of being
   less efficient (specially when searching for neighbours).

   This implemetation follows the paper
     - A. Yershova and S. M. LaValle, Motion Planning for Highly
       Costrained Spaces, Romoco, 2009.
     .

   \sa TKDtree cuik-kdtree.h
*/

/** 
   \brief Method to determine the split point.

   If set to 1, the split point when building the kd-tree is
   computing using the median. Otherwise, we use the
   mean as an approximation to the median.

   Using the median is supposed to produce more balanced
   trees.
*/
#define TRUE_MEDIAN 0

/** 
   \brief Auxiliary macro to print the kd-tree

   Indents accorging to the node height.

   \param f The file where to print the kd-tree.
   \param h The node height.
*/
#define KDTREE_PRINT_INDENT(f,k,indent) for(k=0;k<indent;k++) fprintf(f,"  ");

/** 
   \brief Initializes a kd-tree from a set of points.

   Kd-tree constructor.

   \param ambient The ambient box for the points.
   \param topology The topology (Real,Circle) of each dimension. Might be NULL 
                   if the topology is Real for all dimensions. 
   \param nd Dimensionality of each point. 
   \param m The maximum number of points in the leaves.
   \param r The expansion factor to use in the sampling.
   \param n The number of points.
   \param points The points.
   \param ids The identifiers of the points.

   \return The kd-tree.
*/
TKDtree *BuildKDTree(Trectangle *ambient,unsigned int *topology,
		     unsigned int nd,unsigned int m,double r,
		     unsigned int n,double **points,unsigned int *ids);

/** 
   \brief Searches the nearest point on a KD-tree.

   Auxiliary funciton for \ref NearestNeighbour.

   \param dr2 Squared distance from the query point to the 
              ambient box in the root of the kd-tree.
   \param p The query point.
   \param d2 Squared distanace to the closest point so far.
   \param id The identifier associated with the nearest-neighbour so far.
   \param kdt The KD-tree to query.

*/
void SearchKDtree(double dr2,double *p,double *d2,
		  unsigned int *id,TKDtree *kdt);

/** 
   \brief Searches the near points on a KD-tree.

   Auxiliary funciton for \ref NeighboursInBall.

   \param dr2 Squared distance from the query point to the 
              ambient box in the root of the kd-tree.
   \param p The query point.
   \param r2 Squared search radius.
   \param n Number of neighbours so far.
   \param m Space for the neighbours allocated so far.
   \param id Identifiers neighbours so far.
   \param kdt The KD-tree to query.
*/
void SearchInBallKDtree(double dr2,double *p,double r2,
			unsigned int *n,unsigned int *m,
			unsigned int **id,TKDtree *kdt);

#if (PLOTKDT)
  /** 
    \brief Plots kd-tree nodes.

    For leaves, plots the nodes and the sampling area, depending on 
    the mode.

    For internal nodes, recursively plots the sub-trees.

    \param f The file where to store the plot.
    \param Plot mode. 1 to plot the nodes. 2 to plot the sampling area,
           4 to plot the bounding box, 8 to plot the ambient box.
	   The plots can be combined.
    \param kdt The kdt node to plot.
  */
  void PlotKDtreeNode(FILE *f,unsigned int mode, TKDtree *kdt);
#endif
/************************************************************************/
/************************************************************************/
/************************************************************************/
TKDtree *BuildKDTree(Trectangle *ambient,unsigned int *topology,
		     unsigned int nd,unsigned int m,double r,
		     unsigned int n,double **points,unsigned int *ids)
{
  TKDtree *kdt;

  NEW(kdt,1,TKDtree);

  kdt->nd=nd;
  kdt->n=n;
  kdt->m=m;
  kdt->r=r;
  CopyRectangle(&(kdt->ambient),ambient);
  if (topology==NULL)
    kdt->topology=NULL;
  else
    {
      NEW(kdt->topology,kdt->nd,unsigned int);
      memcpy(kdt->topology,topology,sizeof(unsigned int)*kdt->nd);
    }

  if (n<m)
    {
      unsigned int i;

      kdt->isLeaf=TRUE;

      kdt->height=1;

      NEW(kdt->p,kdt->m+1,double *);  /* +1 is useful when adding points */
      NEW(kdt->id,kdt->m+1,unsigned int);

      if (n==0)
	kdt->volume=0.0;
      else
	{
	  for(i=0;i<n;i++)
	    {
	      kdt->p[i]=points[i];
	      kdt->id[i]=ids[i];
	      if (i==0)
		InitRectangleFromPoint(kdt->nd,points[i],&(kdt->boundingBox));
	      else
		ExpandRectangle(points[i],&(kdt->boundingBox));
	    }
	  CopyRectangle(&(kdt->samplingBox),&(kdt->boundingBox)); /* Just to init samplingBox */
	  kdt->volume=EnlargeRectangleWithLimits(r,
						 &(kdt->ambient),
						 &(kdt->boundingBox),
						 &(kdt->samplingBox));
	}
      kdt->nextLeaf=NULL;

      #ifndef NDEBUG
        /* Reset the internal node info */
        kdt->splitDim=NO_UINT;
	kdt->splitValue=INF;
	kdt->left=NULL;
	kdt->right=NULL;
	kdt->leftLeaf=NULL;
	kdt->rightLeaf=NULL;
      #endif
    }
  else
    {
      signed int i,j,it;
      double *ps;
      unsigned int is;
      TKDtree *kdtLeft,*kdtRight;
      double u,l;

      kdt->isLeaf=FALSE;

      /* determine the split dimension and split point */
      kdt->splitDim=GetRectangleSplitDim(&(kdt->ambient));
      it=0;
      do {
	#if (TRUE_MEDIAN)
	{
	  unsigned int med,right,left;
	  double val;

	  /* We adapt the quickselect algorithm to find the median in O(n) */
	  med=(unsigned int)n/2; //position of the median
	  left=0;
	  right=n;
	  i=0; // do not start the loop if r==l==med
	  while (i!=med)
	    {
	      //we use the leftmost value as a pivot 
	      //(can be any index betwee 'left' and 'right')
	      val=points[left][kdt->splitDim]; 
	    
	      /*
	      // this is only necessary if we pick a pivot value 
	      // different from the left-most one
	      if (pivot!=left)
	      {
	      //Save the pivot value in the leftmost position
	      SWAP(points[left],points[i],ps);
	      SWAP(ids[left],ids[i],is);
	      }
	      */

	      // for the rest of values, find the first value
	      // larger than the pivot value
	      i=left+1;
	      while ((i<right)&&(points[i][kdt->splitDim]<val))
		i++;
	      // classify the values after 'i'
	      j=i+1;
	      while (j<right)
		{
		  if (points[j][kdt->splitDim]<val)
		    {
		      SWAP(points[i],points[j],ps);
		      SWAP(ids[i],ids[j],is); 
		      i++;
		    }
		  j++;
		}
	      if (left!=i-1)
		{
		  // insert the pivot just before the 1st value larger than it
		  SWAP(points[left],points[i-1],ps);
		  SWAP(ids[left],ids[i-1],is); 
		}
	      if (i<med)
		left=i;
	      else
		{
		  if (i>med)
		    right=i;
		}
	    }
	  kdt->splitValue=points[i][kdt->splitDim];
	}
        #else
          //approximate the median by the mean
          kdt->splitValue=0.0;
	  for(i=0;i<n;i++)
	    kdt->splitValue+=points[i][kdt->splitDim];
	  kdt->splitValue/=(double)n;


	  //split the points in two sub-sets according to the split point
	  i=0; 
	  j=n-1;
	  do {
	    while ((i<n)&&(points[i][kdt->splitDim]<kdt->splitValue))
	      i++;
	    while ((j>=0)&&(points[j][kdt->splitDim]>=kdt->splitValue))
	      j--;
	    if ((i<n)&&(j>=0)&&(i<j)&&
		(points[i][kdt->splitDim]>=kdt->splitValue)&&
		(points[j][kdt->splitDim]<kdt->splitValue))
	      {
		SWAP(points[i],points[j],ps);
		SWAP(ids[i],ids[j],is); 
	      }
	  } while (i<j);
        #endif
	
	/* at this point we have the split value and the vector of points is
	   sorted with 'i' elements with values below the split value and
	   n-i elements with values above it. */
	if ((i==0)||(i==n))
	  {
	    /* if we picked a split dimension where all points have the same
	       value, just pick the next one. */
	    kdt->splitDim++;
	    kdt->splitDim%=kdt->nd;
	    /* but there are a limited number of dimensions :) */
	    it++;
	    assert(it<kdt->nd);
	  }
      }	 while((i==0)||(i==n));

      /* Recursively build kd-trees */
	    
      GetRectangleLimits(kdt->splitDim,&l,&u,&(kdt->ambient));

      SetRectangleUpperLimit(kdt->splitDim,kdt->splitValue,&(kdt->ambient));
      kdtLeft=BuildKDTree(&(kdt->ambient),topology,nd,m,r,i,points,ids);

      SetRectangleLimits(kdt->splitDim,kdt->splitValue,u,&(kdt->ambient));
      kdtRight=BuildKDTree(&(kdt->ambient),topology,nd,m,r,n-i,&(points[i]),&(ids[i]));

      SetRectangleLowerLimit(kdt->splitDim,l,&(kdt->ambient));
      
      /* Assemble the sub-trees */ 
      kdt->left=(void *)kdtLeft;
      kdt->right=(void *)kdtRight;

      kdt->leftLeaf=(kdtLeft->isLeaf?(void*)kdtLeft:kdtLeft->leftLeaf);
      kdt->rightLeaf=(kdtRight->isLeaf?(void*)kdtRight:kdtRight->rightLeaf);

      /* Update the smapling area */
      kdt->volume=kdtLeft->volume+kdtRight->volume;
      
      /* Update the height */
      kdt->height=1+(kdtLeft->height>kdtRight->height?
		     kdtLeft->height:
		     kdtRight->height);

      /* Chain the sub-tree leaves */
      if (kdtLeft->isLeaf)
	{
	  if (kdtRight->isLeaf)
	    kdtLeft->nextLeaf=(void *)kdtRight;
	  else
	    kdtLeft->nextLeaf=kdtRight->leftLeaf;
	}
      else
	{
	  if (kdtRight->isLeaf)
	    KDTPTR(kdtLeft->rightLeaf)->nextLeaf=(void *)kdtRight;
	  else
	    KDTPTR(kdtLeft->rightLeaf)->nextLeaf=kdtRight->leftLeaf;
	}

      #ifndef NDEBUG
        /* Reset the leaf info */
        kdt->nextLeaf=NULL;
	kdt->p=NULL;
	kdt->id=NULL;
      #endif
    }

  return(kdt);
}

void SearchKDtree(double dr2,double *p,double *d2,
		  unsigned int *id,TKDtree *kdt)
{
  if (kdt->isLeaf)
    {
      if (SquaredDistanceToRectangle(*d2,p,kdt->topology,&(kdt->boundingBox))<*d2)
	{
	  unsigned int i;
	  double dp;
      
	  for(i=0;i<kdt->n;i++)
	    {
	      dp=VectorSquaredDistanceTopologyMin(*d2,kdt->nd,kdt->topology,p,kdt->p[i]);
	      if (dp<*d2)
		{
		  *d2=dp;
		  *id=kdt->id[i];
		}
	    }
	}
    }
  else
    {
      TKDtree *t1,*t2;
      double dt0,dt1,dt2,d,v;

      v=p[kdt->splitDim];
      if (v<=kdt->splitValue)
	{
	  /* The nearest point is most likely in the left branch */
	  t1=KDTPTR(kdt->left);
	  t2=KDTPTR(kdt->right);
	}
      else
	{
	  /* The nearest point is most likely in the left branch */
	  t1=KDTPTR(kdt->right);
	  t2=KDTPTR(kdt->left);
	}

      dt0=dr2-SquaredDistanceToRectangleDimension(kdt->splitDim,v,kdt->topology,
						  &(kdt->ambient));

      dt1=SquaredDistanceToRectangleDimension(kdt->splitDim,v,kdt->topology,
					      &(t1->ambient));
      d=dt0+dt1;
      if (d<*d2)
	SearchKDtree(d,p,d2,id,t1);

      dt2=SquaredDistanceToRectangleDimension(kdt->splitDim,v,kdt->topology,
					      &(t2->ambient));
      d=dt0+dt2;
      if (d<*d2)
	SearchKDtree(d,p,d2,id,t2);
    }
}

void SearchInBallKDtree(double dr2,double *p,double r2,
			unsigned int *n,unsigned int *m,
			unsigned int **id,TKDtree *kdt)
{
  if (kdt->isLeaf)
    {
      if (SquaredDistanceToRectangle(r2,p,kdt->topology,&(kdt->boundingBox))<r2)
	{
	  unsigned int i;
	  
	  for(i=0;i<kdt->n;i++)
	    {
	      if (VectorSquaredDistanceTopologyMin(r2,kdt->nd,kdt->topology,p,kdt->p[i])<r2)
		{
		  if (*n==*m)
		    MEM_DUP(*id,*m,unsigned int);
		  (*id)[*n]=kdt->id[i];
		  (*n)++;
		}
	    }
	}
    }
  else
    {
      TKDtree *t;
      double d,dt,v;

      v=p[kdt->splitDim];
      dt=dr2-SquaredDistanceToRectangleDimension(kdt->splitDim,v,kdt->topology,
						 &(kdt->ambient));
       
      t=KDTPTR(kdt->left);
      d=dt+SquaredDistanceToRectangleDimension(kdt->splitDim,v,kdt->topology,
					       &(t->ambient));
      if (d<r2)
	SearchInBallKDtree(d,p,r2,n,m,id,t);
       
      t=KDTPTR(kdt->right);
      d=dt+SquaredDistanceToRectangleDimension(kdt->splitDim,v,kdt->topology,
					       &(t->ambient));
      if (d<r2)
	SearchInBallKDtree(d,p,r2,n,m,id,t);
    }
}

#if (PLOTKDT)
  void PlotKDtreeNode(FILE *f,unsigned int mode, TKDtree *kdt)
  {
    if (kdt->isLeaf)
      {
	unsigned int i;
	double lx,ux,ly,uy,lz,uz;

	/* Plot the points (in red) */
	if (mode&1)
	  {
	    for(i=0;i<kdt->n;i++)
	      {
		fprintf(f,"{VECT 1 2 1 \n   2 \n   1\n");
		fprintf(f,"   %f %f %f\n",kdt->p[i][0]-0.01,kdt->p[i][1],kdt->p[i][2]);
		fprintf(f,"   %f %f %f\n",kdt->p[i][0]+0.01,kdt->p[i][1],kdt->p[i][2]);
		fprintf(f," 1 0 0 0}\n");
		fprintf(f,"{VECT 1 2 1 \n   2 \n   1\n");
		fprintf(f,"   %f %f %f\n",kdt->p[i][0],kdt->p[i][1]-0.01,kdt->p[i][2]);
		fprintf(f,"   %f %f %f\n",kdt->p[i][0],kdt->p[i][1]+0.01,kdt->p[i][2]);
		fprintf(f," 1 0 0 0}\n");
		fprintf(f,"{VECT 1 2 1 \n   2 \n   1\n");
		fprintf(f,"   %f %f %f\n",kdt->p[i][0],kdt->p[i][1],kdt->p[i][2]-0.01);
		fprintf(f,"   %f %f %f\n",kdt->p[i][0],kdt->p[i][1],kdt->p[i][2]+0.01);
		fprintf(f," 1 0 0 0}\n");
	      }
	  }

	if (mode&12)
	  {
	    Trectangle *r;

	    if (mode&2)
	      r=&(kdt->samplingBox);  /* Plot the sampling box */
	    else
	      {
		if (mode&4)
		  r=&(kdt->boundingBox);   /* Plot the bouding box */
		else
		  r=&(kdt->ambient);    /* Plot the ambient box */
	      }
	    
	    GetRectangleLimits(0,&lx,&ux,r);
	    GetRectangleLimits(1,&ly,&uy,r);
	    GetRectangleLimits(2,&lz,&uz,r);

	    fprintf(f,"{OFF 8 6 1\n");
	    fprintf(f,"%f %f %f\n",lx,ly,lz);
	    fprintf(f,"%f %f %f\n",lx,uy,lz);
	    fprintf(f,"%f %f %f\n",ux,uy,lz);
	    fprintf(f,"%f %f %f\n",ux,ly,lz);
	
	    fprintf(f,"%f %f %f\n",lx,ly,uz);
	    fprintf(f,"%f %f %f\n",lx,uy,uz);
	    fprintf(f,"%f %f %f\n",ux,uy,uz);
	    fprintf(f,"%f %f %f\n",ux,ly,uz);
	
	    /*faces*/
	    fprintf(f,"4 3 2 1 0\n");
	    fprintf(f,"4 4 5 6 7\n");
	    fprintf(f,"4 0 1 5 4\n"); 
	    fprintf(f,"4 2 3 7 6\n"); 
	    fprintf(f,"4 3 0 4 7\n");
	    fprintf(f,"4 1 2 6 5}\n"); 
	  }
      }
    else
      {
	PlotKDtreeNode(f,mode,KDTPTR(kdt->left));
	PlotKDtreeNode(f,mode,KDTPTR(kdt->right));
      }
  }
#endif

/************************************************************************/
/************************************************************************/
/************************************************************************/

TKDtree *InitKDTree(Trectangle *ambient,unsigned int *topology,unsigned int m,double r,
		    unsigned int n,double **points,unsigned int *ids)
{
  unsigned int nd;

  assert(m>=2);

  /* Correct the ambient box using the topology */
  nd=GetRectangleDim(ambient);
  if (topology!=NULL)
    {
      unsigned int i;
      double l,u;

      for(i=0;i<nd;i++)
	{
	  if (topology[i]==TOPOLOGY_S)
	    {
	      GetRectangleLimits(i,&l,&u,ambient);

	      assert(l<=u);

	      /* If the given range is close to -Pi / Pi but
	         for rounding errors, we fix the values to avoid
	         numerical errors. */
	      if (fabs(l+M_PI)<ANGLE_ACCURACY)
		l=-M_PI;
	      if (fabs(u-M_PI)<ANGLE_ACCURACY)
		u=M_PI;
	      
	      SetRectangleLimits(i,l,u,ambient);
	    }
	}
    }

  return(BuildKDTree(ambient,topology,nd,m,r,n,points,ids));
}

void AddPoint2KDtree(double *point,unsigned int id,TKDtree **kdt)
{
  TKDtree *t;

  t=*kdt;

  if (t->isLeaf)
    {
      #ifndef NDEBUG
	/* Check for the presense of repeated points in the tree */
	unsigned int j;
	double e;
	
	for(j=0;j<t->n;j++)
	  {
	    e=VectorSquaredDistanceTopologyMin(INF,t->nd,t->topology,t->p[j],point);
	    if (e<1e-12) /*threshold 1e-6*1e-6*/
	      {
		fprintf(stderr,"Repeated point in kd-tree\n");
		exit(0);
	      }
	  }
      #endif
      
      t->p[t->n]=point;
      t->id[t->n]=id;
      t->n++;
      if (t->n>t->m)
	{
	  TKDtree *nt;

	  nt=BuildKDTree(&(t->ambient),t->topology,t->nd,t->m,t->r,t->n,t->p,t->id);
	  DeleteKDtree(t);
	  *kdt=nt;
	}
      else
	{
	  if (t->n==1)
	    {
	      /* adding the first point to an empty tree/leaf */
	      InitRectangleFromPoint(t->nd,point,&(t->boundingBox));
	      CopyRectangle(&(t->samplingBox),&(t->boundingBox));  /* just to allocate space */
	    }
	  else
	    ExpandRectangle(point,&(t->boundingBox));
	  t->volume=EnlargeRectangleWithLimits(t->r,
					       &(t->ambient),
					       &(t->boundingBox),
					       &(t->samplingBox));
	}
    }
  else
    {
      TKDtree *kdtLeft,*kdtRight;

      if (point[t->splitDim]<=t->splitValue)
	AddPoint2KDtree(point,id,(TKDtree **)&(t->left));
      else
	AddPoint2KDtree(point,id,(TKDtree **)&(t->right));

      t->n++;

      kdtLeft=KDTPTR(t->left);
      kdtRight=KDTPTR(t->right);

      /* One of the sub-trees changed -> update the summary at this level */
      
      /* The chain of leaves must be updated even if we re-balance.
         Before re-balancing, we need a correct chain of leaves. */

      t->leftLeaf=(kdtLeft->isLeaf?(void*)kdtLeft:kdtLeft->leftLeaf);
      t->rightLeaf=(kdtRight->isLeaf?(void*)kdtRight:kdtRight->rightLeaf);

      /* Chain the sub-tree leaves */
      if (kdtLeft->isLeaf)
	{
	  if (kdtRight->isLeaf)
	    kdtLeft->nextLeaf=(void *)kdtRight;
	  else
	    kdtLeft->nextLeaf=kdtRight->leftLeaf;
	}
      else
	{
	  if (kdtRight->isLeaf)
	    KDTPTR(kdtLeft->rightLeaf)->nextLeaf=(void *)kdtRight;
	  else
	    KDTPTR(kdtLeft->rightLeaf)->nextLeaf=kdtRight->leftLeaf;
	}

      /* If necessary, rebalance the tree -> re-build the tree */
      if ((KDTPTR(t->left)->height>KDTPTR(t->right)->height*2)||
	  (KDTPTR(t->right)->height>KDTPTR(t->left)->height*2))
	{
	  double **p;
	  unsigned int *id;
	  TKDtree *l,*nt;
	  unsigned int i,j;

	  /* Collect all the points in the leaves of the tree to rebuild */
	  NEW(p,t->n,double *);
	  NEW(id,t->n,unsigned int);
	  l=KDTPTR(t->leftLeaf);
	  i=0;
	  while(i<t->n)
	    {
	      for(j=0;j<l->n;j++)
		{
		  p[i]=l->p[j];
		  id[i]=l->id[j];
		  i++;
		}
	      l=KDTPTR(l->nextLeaf);
	    }
	  
	  /* Re-build the tree */
	  nt=BuildKDTree(&(t->ambient),t->topology,t->nd,t->m,t->r,t->n,p,id);
	  DeleteKDtree(t);
	  *kdt=nt;

	  free(p);
	  free(id);
	}
      else
	{
	  /* Some "summary" information is only valid if we do not re-balance */
	  t->volume=kdtLeft->volume+kdtRight->volume;	  
	  t->height=1+(kdtLeft->height>kdtRight->height?
		       kdtLeft->height:kdtRight->height);
	}
    }
}

unsigned int NearestNeighbour(double *point,TKDtree *kdt)
{
  unsigned int id;
  double d;

  assert(kdt->n>0);

  id=NO_UINT;
  d=INF;
  SearchKDtree(0.0,point,&d,&id,kdt);

  assert(id!=NO_UINT);

  return(id);
}

void NeighboursInBall(double *point,double r,
		      unsigned int *n,unsigned int **ids,TKDtree *kdt)
{
  unsigned int m;

  assert(kdt->n>0);

  m=100;
  NEW(*ids,m,unsigned int);
  *n=0;
 
  SearchInBallKDtree(0.0,point,r*r,n,&m,ids,kdt);

  if (*n==0)
    free(*ids);
}

void SampleInKDtree(double *point,TKDtree *kdt)
{
  if (kdt->isLeaf)
    RandomPointInRectangle(point,&(kdt->samplingBox));
  else
    {
      double r;

      r=randomDouble()*kdt->volume;
      
      if (r<KDTPTR(kdt->left)->volume)
	SampleInKDtree(point,KDTPTR(kdt->left));
      else
	SampleInKDtree(point,KDTPTR(kdt->right));
    }
}

double KDtreeSamplingExpansion(TKDtree *kdt)
{
  return(kdt->r);
}

void PrintKDtree(FILE *f,unsigned int indent,TKDtree *kdt)
{
  unsigned int k;

  if (kdt->isLeaf)
    {
      unsigned int i,j;

      KDTREE_PRINT_INDENT(f,k,indent);PrintRectangle(f,&(kdt->boundingBox));fprintf(f,"\n");
      for(i=0;i<kdt->n;i++)
	{
	  KDTREE_PRINT_INDENT(f,k,indent);
	  fprintf(f,"%u: [",kdt->id[i]);
	  for(j=0;j<kdt->nd;j++)
	    fprintf(f,"%f ",kdt->p[i][j]);
	  fprintf(f,"]\n");
	}
    }
  else
    {
      KDTREE_PRINT_INDENT(f,k,indent);fprintf(f,"Split dim: %u\n",kdt->splitDim);
      KDTREE_PRINT_INDENT(f,k,indent);fprintf(f,"Split val: %f\n",kdt->splitValue);
      KDTREE_PRINT_INDENT(f,k,indent);fprintf(f,"N. points: %u\n",kdt->n);
      KDTREE_PRINT_INDENT(f,k,indent);fprintf(f,"Left tree:\n");
      PrintKDtree(f,indent+1,KDTPTR(kdt->left));
      KDTREE_PRINT_INDENT(f,k,indent);fprintf(f,"Right tree:\n");
      PrintKDtree(f,indent+1,KDTPTR(kdt->right));
    }
}

#if (PLOTKDT)
  void PlotKDtree(char *fname,boolean sphere,TKDtree *kdt)
  {
    FILE *f;
    
    assert(kdt->nd==3);

    f=fopen(fname,"w");
    assert(f);

    fprintf(f,"(progn (normalization World none)\n(bbox-draw World no))\n");
    
    if (sphere)
      fprintf(f,"(geometry ball {LIST {SPHERE 1 0 0 0}})\n");
    
    //points
    fprintf(f,"(geometry kdtree-nodes {LIST \n"); 
      PlotKDtreeNode(f,1,kdt);
    fprintf(f,"})\n"); 

    //sampling volume
    fprintf(f,"(geometry kdtree-sampling {LIST \n"); 
      PlotKDtreeNode(f,2,kdt);
    fprintf(f,"})\n");

    /*
    //bouding box
    fprintf(f,"(geometry kdtree-bounding {LIST \n"); 
      PlotKDtreeNode(f,4,kdt);
    fprintf(f,"})\n");
    */

    fclose(f);
  }
#endif

void DeleteKDtree(TKDtree *kdt)
{
  if (kdt->isLeaf)
    {
      free(kdt->p);
      free(kdt->id);
      if (kdt->n>0)
	{
	  DeleteRectangle(&(kdt->boundingBox));
	  DeleteRectangle(&(kdt->samplingBox));
	}
    }
  else
    {
      DeleteKDtree(KDTPTR(kdt->left));
      DeleteKDtree(KDTPTR(kdt->right));
    }
  DeleteRectangle(&(kdt->ambient));
  if (kdt->topology!=NULL)
    free(kdt->topology);
  free(kdt);
}
