#ifndef CUIKKDTREEH
#define CUIKKDTREEH

#include "definitions.h"
#include "rectangle.h"

/** 
   \file cuik-kdtree.h

   \brief Headers of the kd-tree implementation.

   Headers of the kd-tree implementation.
   
   \sa TKDtree cuik-kdtree.c
*/

/** 
   \brief Flag used for debug.

   When set, it generates a file of the kd-tree that can be 
   visualized with geomview.

   This only works for kd-trees defined in 3d and it is 
   more useful when the points are defined on a manifold
   (see \ref POINTS_ON_SPHERE).
*/
#ifdef NDEBUG
  #define PLOTKDT 0
#else
  #define PLOTKDT 1
#endif

/** 
   \brief Void pointer to KDtree pointer

   Casts a void pointer into a KDtree pointer.
*/
#define KDTPTR(a) ((TKDtree *)(a))

/** 
    \brief A KD-tree.

    This structure is used to represent a kd-tree or any kd-sub-tree, including
    leaves.

    For simplicity reasons, we us a recursive implementation of most of the
    methods for the kd-tree.

    Note that our implementation of the kd-tree does not include the points but
    only pointer to them.
*/
typedef struct {
  boolean isLeaf;           /**< TRUE if the sub-tree is a leaf */
  
  /* for all sub-trees */
  unsigned int nd;          /**< Dimension of the ponits. */
  unsigned int n;           /**< Number of points in this sub-tree. */
  unsigned int m;           /**< Maximum number of points in the leaves. */
  double r;                 /**< The expansion factor to use in the sampling. */
  Trectangle ambient;       /**< Box defining the ambient space for all points. */
  unsigned int *topology;   /**< Topology (Real,Circle) for each dimension. */
  unsigned int height;      /**< Height of this sub-tree. */
  double volume;            /**< Volume of the sampling area. */

  /* for internal nodes */
  unsigned int splitDim;    /**< Dimension used to split the points. */
  double splitValue;        /**< Value used to split the points. */
  void *left;               /**< Left sub-tree. */
  void *right;              /**< Rght sub-tree. */
  void *leftLeaf;           /**< Left-most leaf in the sub-tree. */
  void *rightLeaf;          /**< Right-most leaf inthe sub-tree. */

  /* for leaves */
  void *nextLeaf;           /**< Next leaf in the tree. */
  double **p;               /**< Points in the leaf. */
  unsigned int *id;         /**< Identifiers of the points in the leaf. */
  Trectangle boundingBox;   /**< Actual bouding box for the points. */
  Trectangle samplingBox;   /**< Sampling box (the bounding box enlarged by 'r'). */

} TKDtree;

/** 
   \brief Initializes a kd-tree from a set of points.

   Kd-tree constructor.

   \param ambient The ambient box for the points.
   \param topology The topology (Real,Circle) of each dimension. Might be NULL if the
                   topology is real for all dimensions.
   \param m The maximum number of points in the leaves.
   \param r The expansion factor to use in the sampling.
   \param n The number of points.
   \param points The points.
   \param ids The identifiers of the points.

   \return The kd-tree.
*/
TKDtree *InitKDTree(Trectangle *ambient,unsigned int *topology,
		    unsigned int m,double r,unsigned int n,
		    double **points,unsigned int *ids);

/** 
   \brief Adds a point to a kd-tree.

   Adds a point to a kd-tree.

   \param point The point to add to the kd-tree.
   \param id The identifier associated with the point.
   \param kdt The kd-tree to update.
*/
void AddPoint2KDtree(double *point,unsigned int id,TKDtree **kdt);

/** 
   \brief Determines the nearest-neighbour for a point.

   Uses the kd-tree to find the nearest-neighbour of a given point.
   
   \param point The query point.
   \param kdt The KD-tree to query.

   \return The identifier associated with the nearest-neighbour.
*/
unsigned int NearestNeighbour(double *point,TKDtree *kdt);

/** 
   \brief Neighbours in a ball
   
   Determines all the points in a given ball.

   \param point The center of the query ball.
   \param r The radius of the query ball.
   \param n The nuber of output points.
   \param ids The identifiers for the output points.
   \param kdt The KD-tree to query. 
*/
void NeighboursInBall(double *point,double r,
		      unsigned int *n,unsigned int **ids,TKDtree *kdt);

/** 
   \brief Generates a random sample.

   Geneates a random sample in the volume covered by the kd-tree expanded by a given
   factor ('r' given in the kd-tree construction) with uniform distribution.

   \param point The sampled point.
   \param kdt The kd-tree

   \sa InitKDTree
*/
void SampleInKDtree(double *point,TKDtree *kdt);

/** 
    \brief Returns the sampling expansion used in the kd-tree.

    Returns the sampling expansion used in the kd-tree.
    
    \param kdt The kd-tree to query.

    \return The sampling expansion of the kd-tree.
*/
double KDtreeSamplingExpansion(TKDtree *kdt);

/** 
   \brief Prints a kd-tree

   Prints a kd-tree. Used for debug.

   \param f The file where to print the kd-tree. Typically stdout or stderr.
   \param indent Indent level for the first level of the tree. The rest of levels
                 are progressively indented.
   \param kdt The kd-tree to print.
*/
void PrintKDtree(FILE *f,unsigned int indent,TKDtree *kdt);

#if (PLOTKDT)
  /** 
    \brief Plots a kd-tree.

    Plots the kd-tree points and the sampling areas bounding these points.
    
    This is used for debug and it only works for points in 3D.

    The generated file con be visualized with geomview.

    \param fname Name of the file to generate.
    \param sphere TRUE if the points are defined on a sphere.
    \param kdt The kd-tree to plot.
  */
  void PlotKDtree(char *fname,boolean sphere,TKDtree *kdt);
#endif

/** 
   \brief Deletes a KD-tree
   
   Destructor
*/
void DeleteKDtree(TKDtree *kdt);

#endif
