#ifndef DEFINITIONSH
#define DEFINITIONSH

/** 
   \file definitions.h

   \brief Definition of constants and macros used in several parts of the library.

   Definition of constants and macros used in several parts of the cuik-kdtree
   library.
*/

#include <stdlib.h> /*The NEW macros call malloc,..*/
#include <math.h>

/************************************************************************************/
/************************************************************************************/

#ifndef M_PI
  /** \brief Pi.

      An approximation of Pi
  */
  #define M_PI 3.14159265358979323846
#endif

#ifndef M_2PI
  /** \brief 2*Pi.

      Two times \ref M_PI "Pi".
  */
  #define M_2PI (2.0*M_PI)
#endif


/************************************************************************************/
/************************************************************************************/
#ifndef ANGLE_ACCURACY
  /** 
      \brief Accuracy used to compare angles.
      
      Two angular values are the same if they are closer than this value.
  */
  #define ANGLE_ACCURACY 1e-4
#endif

/************************************************************************************/
/************************************************************************************/

#ifndef SWAP
  /** 
      \brief Swaps two values.
      
      Swaps two values using an intermediate storage.
      
      \param a First value to swap.
      \param b Second value to swap.
      \param c Intermediate storage for the swap.
  */
  #define SWAP(a,b,c) {c=a;a=b;b=c;}
#endif

/************************************************************************************/
/************************************************************************************/

#ifndef PI2PI
  /** \brief Forces an angle go be in [-pi,pi].

      Forces an angle go be in [-pi,pi].
      This is intended to be used with a variable as a parameter
      (using an expression would be ineficient).
      
      \param a The variable to set within the [-pi,pi] range.
  */
  #define PI2PI(a) { if ((a<-M_PI)||(a>M_PI)) { while(a>M_PI) a-=M_2PI; while (a<-M_PI) a+=M_2PI; }}
#endif

/************************************************************************************/
/************************************************************************************/

#ifndef NEW
  /** \brief Allocates memory space.
      
      Allocates memory space for \e _n elements of type \e _type and assigns the pointer to \e _var.
      If the memory can not be allocated, stops the execution.
  */
  #define NEW(_var,_n,_type) {(_var)=(_type *)malloc(sizeof(_type)*(_n)); \
                               if (!(_var)) exit(EXIT_FAILURE);}
#endif

#ifndef NEWZ
  /** \brief Allocates memory space and initializes to zero.
      
      Allocates memory space for \e _n elements of type \e _type and assigns the pointer to \e _var.
      If the memory can not be allocated, stops the execution.
      The allocated memory is set to 0.
  */
#define NEWZ(_var,_n,_type) {(_var)=(_type *)calloc(_n,sizeof(_type));		\
                               if (!(_var)) exit(EXIT_FAILURE);}
#endif

#ifndef MEM_EXPAND
  /** \brief Expands a previously allocated memory space.
      
      Expands a previously reserved memory space
      so that it can accomodate \e _n elements of type \e _type  assigns the pointer to \e _var.
      The previously allocated space is pointed by \e _var.
      If the memory can not be expanded, stops the execution.
  */
  #define MEM_EXPAND(_var,_n,_type) {(_var)=(_type *)realloc((void *)(_var),sizeof(_type)*(_n)); \
                                     if (!(_var)) exit(EXIT_FAILURE);}
#endif

#ifndef MEM_DUP
  /** \brief Duplicates a previously allocated memory space.

      Dublicates a previously reserved memory space
      that accomodates \e _n elements of type \e _type.
      After the \e duplication, the pointer \e _var takes to a memory space
      that can accomodate 2*\e _n elements and the value of variable \e _n 
      is also doubled.
  */
  #define MEM_DUP(_var,_n,_type) {(_n)*=2;MEM_EXPAND(_var,_n,_type);}
#endif

/************************************************************************************/
/************************************************************************************/
#ifndef NO_UINT
  /** \brief Used to denote an identifier that has not been initialized.

      Used to denote an identifier that has not been initialized.
  */
  #define NO_UINT ((unsigned int)(-1))
#endif
 
#ifndef INF
  /** \brief Infinite.
    
      Infinite (a large possitive number).
  */
  #define INF 1e200
#endif

/************************************************************************************/
/************************************************************************************/
#ifndef boolean
  /** 
      \brief Boolean type.
      
      Type with two possible values: TRUE,FALSE.
  */
  typedef unsigned int boolean; 
#endif

#ifndef TRUE
  /** 
     \brief TRUE

     TRUE   
  */
  #define TRUE  1
#endif

#ifndef FALSE
  /** 
     \brief FALSE

     FALSE   
  */
  #define FALSE 0
#endif

/************************************************************************************/
/************************************************************************************/
#ifndef TOPOLOGY_R
  /** \brief One of the possible topologies.
    
      The range of the variables is defined over the reals.
  */
  #define TOPOLOGY_R 1
#endif

#ifndef TOPOLOGY_S
  /** \brief One of the possible topologies.
    
      The range of the variables is defined over a circle.
  */
  #define TOPOLOGY_S 2
#endif

#endif
