#ifndef RECTANGLEH
#define RECTANGLEH

#include "definitions.h"

#include <stdio.h>


/** \file rectangle.h
    \brief Definition of the Trectangle type and the associated functions.

    Definition of the Trectangle type and the associated functions.

    \sa Trectangle, rectangle.c.
*/



/** 
 \brief A rectangle.

 A rectangle is a <em>n</em>-dimensional orthotope, i.e., a vector of ranges.

 \sa rectangle.h
*/
typedef struct {
  unsigned int n; /**< Number of dimensions of the rectangle. */
  double *l;      /**< Lower limit for the ranges defining the rectanble. */
  double *u;      /**< Upper limit for the ranges defining the rectanble. */
} Trectangle;

/** 
    \brief Initializes a rectangle.

    Creates a new rectangle with the given parameters.

    \param dim Dimension of the new rectangle.
    \param l Lower limits.
    \param u Upper limits.
    \param b The rectangle to initialize.
*/
void InitRectangle(unsigned int dim,double *l,double *u,Trectangle *b);

/** 
    \brief Initializes a rectangle from a point.
 
    Creates a new rectangle with punctual intervals from a given vector.

    \param dim Dimension of the new rectangle (i.e., number of intervals).
    \param p The point to use to initialize the rectangle.
    \param b The rectangle to initialize.
*/
void InitRectangleFromPoint(unsigned int dim,double *p,Trectangle *b);

/** 
   \brief Returns the dimension of the rectangle.

   Returns the number of ranges defining the rectangle.

   \param b The rectangle to query.

   \return The number of ranges in the rectangle.
*/
unsigned int GetRectangleDim(Trectangle *b);

/** 
    \brief Expands a rectangle so that it includes a given point.

    Expands a rectangle so that it includes a given point.

    \param p The point.
    \param b The rectangle to modify.

    \sa ExpandInterval.
*/
void ExpandRectangle(double *p,Trectangle *b);

/** 
    \brief Enlarges a box remaining in a given limits.

    Adds a border to a box, but without scaping a given limits
    
    \param r Size of the border to add to the box.
    \param limits The maximum limit for the box expansion. This box is assumed
                  to fully include 'bIn'.
    \param bIn The box to enlarge.
    \param bOut The output box.

    \return The volume of the box after updating it.
*/
double EnlargeRectangleWithLimits(double r,Trectangle *limits,
				  Trectangle *bIn,Trectangle *bOut);

/** 
    \brief Rectangle copy operator.

    Copies one rectangle into another. 

    \param b_out Output rectangle. Rectangle constructed by the copy.
    \param b_in Input rectangle. Rectangle used to create b_out.

*/
void CopyRectangle(Trectangle *b_out,Trectangle *b_in);

/** 
   \brief Get the lower limit.

   Gets the lower limit along a given dimension.

   \param i The dimension.
   \param b The rectagle to modify.

   \return The lower limit.
*/
inline double GetRectangleLowerLimit(unsigned int i,Trectangle *b);

/** 
   \brief Get the upper limit.

   Gets the upper limit along a given dimension.

   \param i The dimension.
   \param b The rectagle to modify.

   \return The upper limit.
*/
inline double GetRectangleUpperLimit(unsigned int i,Trectangle *b);

/** 
   \brief Gets the limits of the rectangle along a given dimension.

   Gets the lower and upper limit of a rectangle along a given dimension.
    
   \param i The dimension.
   \param l The lower limit (output).
   \param u The upper limit (output).
   \param b The rectagle to query.
*/
inline void GetRectangleLimits(unsigned int i,double *l,double *u,Trectangle *b);

/** 
   \brief Set the lower limit.

   Set the lower limit along a given dimension.

   \param i The dimension.
   \param l The new lower limit.
   \param b The rectagle to modify.
*/
inline void SetRectangleLowerLimit(unsigned int i,double l,Trectangle *b);

/** 
   \brief Set the upper limit.

   Set the upper limit along a given dimension.

   \param i The dimension.
   \param u The new upper limit.
   \param b The rectagle to modify.
*/
inline void SetRectangleUpperLimit(unsigned int i,double u,Trectangle *b);

/** 
   \brief Changes a rectangle along a given dimension.

   Changes the lower and the upper limit along a given dimension.
    
   \param i The dimension.
   \param l The new lower limit.
   \param u The new upper limit.
   \param b The rectagle to modify.
*/
inline void SetRectangleLimits(unsigned int i,double l,double u,Trectangle *b);

/** 
    \brief Returns the a random point along the selected dimensions.

    Computes a random point along the selected dimensions.

    \param c Space where to store the random point (should be allocated by the caller
             with, at least, space for the point along the selected dimensions).
    \param b The rectangle.
*/
void RandomPointInRectangle(double *c,Trectangle *b);

/** 
   \brief Squared distance from a value to a given rectangle dimension.

   Squared distance of a value to the interval defining a rectangle along a given dimension.

   \param dim The dimension.
   \param p The point.
   \param tp The topology of the space where the rectangle and the point
             are defined.
   \param b The rectangle.

   \return The squared distance from the value to the interval.
*/
double SquaredDistanceToRectangleDimension(unsigned int dim,double p,unsigned int *tp,Trectangle *b);

/** 
   \brief Squared distance from a point to a rectangle.

   Computes the minimum squared distance from a point to a rectangle but considering
   a possible special topology for the input rectangle and the input point.

   \param t2 Threshold. Only poits closer than 't2' to the rectangle are relevant.
             Use INF if all points must be considered. Using a threshold
	     accelerates the computation in some cases.
   \param p The point.
   \param tp The topology of the space where the rectangle and the point
             are defined.
   \param b The rectangle.

   \return The minimum distance between the input parameters.
*/
double SquaredDistanceToRectangle(double t2,double *p,unsigned int *tp,Trectangle *b);

/**     
    \brief Computes the rectangle dimension for which it is better to split the rectangle.

    Computes the rectangle dimension for which it is better to split the rectangle. Right now
    this function is an alias to GetRectangleMaxDim.

    \param b The rectangle.
    
    \return The best split dimension for the given rectangle.

    \sa GetRectangleMaxDim
*/
unsigned int GetRectangleSplitDim(Trectangle *b);

/** 
   \brief Prints a rectangle.

   Prints a rectangle. Used for debug.

   \param f The file where to print the rectangle. Typically stdout or stderr.
   \param b The rectangle to print.
*/
void PrintRectangle(FILE *f,Trectangle *b);

/**     
    \brief Destructor

    Deletes a previously generated rectangle.

    \param b The rectangle to be deleted.
      
*/
void DeleteRectangle(Trectangle *b);

#endif
